import sys
import inspect
import logging

from k3proc import command

logger = logging.getLogger(__name__)

#  Since 3.8 there is a stacklevel argument
ddstack_kwarg = {}
if sys.version_info.major == 3 and sys.version_info.minor >= 8:
    ddstack_kwarg = {"stacklevel": 2}


def dd(*msg):
    """
    Alias to logger.debug()
    """
    msg = [str(x) for x in msg]
    msg = " ".join(msg)
    logger.debug(msg, **ddstack_kwarg)


def ddstack(*msg):
    """
    Log calling stack in logging.DEBUG level.
    """

    if logger.isEnabledFor(logging.DEBUG):
        stack = inspect.stack()[1:]
        for i, (frame, path, ln, func, line, xx) in enumerate(stack):
            #  python -c "xxx" does not have a line
            if line is None:
                line = ""
            else:
                line = line[0].strip()
            logger.debug("stack: %d %s %s", ln, func, line, **ddstack_kwarg)


def cmdf(cmd, *arguments, flag="", **options):
    """
    Alias to k3proc.command(). Behaviors is specified with ``flag``

    Args:
        cmd(str): the path to executable.

        arguments: arguments

        flag(str or list or tuple): str flags.

            - 'x' or ('raise', ): raise CalledProcessError if return code is not 0
            - 't' or ('tty', ): start sub process in a tty.
            - 'n' or ('none', ): if return code is not 0, return None.
            - 'p' or ('pass', ): do not capture stdin, stdout and stderr.
            - 'o' or ('stdout', ): only return stdout in list of str.
            - '0' or ('oneline', ): only return the first line of stdout.

        options: other options pass to k3proc.command().

    Returns:
        str: first line of stdout.
    """
    dd("cmdf:", cmd, arguments, options)
    dd("flag:", flag)
    flag = parse_flag(flag)

    if "raise" in flag:
        options["check"] = True
    if "tty" in flag:
        options["tty"] = True
    if "pass" in flag:
        options["capture"] = False

    code, out, err = command(cmd, *arguments, **options)

    # reaching here means there is no check of exception
    if code != 0 and "none" in flag:
        return None

    out = out.splitlines()
    err = err.splitlines()

    if "stdout" in flag:
        dd("cmdf: out:", out)
        return out

    if "oneline" in flag:
        dd("cmdf: out:", out)
        if len(out) > 0:
            return out[0]
        return ""

    return code, out, err


def cmd0(cmd, *arguments, **options):
    """
    Alias to k3proc.command() with ``check=True``

    Returns:
        str: first line of stdout.
    """
    dd("cmd0:", cmd, arguments, options)
    _, out, _ = cmdx(cmd, *arguments, **options)
    dd("cmd0: out:", out)
    if len(out) > 0:
        return out[0]
    return ""


def cmdout(cmd, *arguments, **options):
    """
    Alias to k3proc.command() with ``check=True``.

    Returns:
        list: stdout in lines of str.
    """

    dd("cmdout:", cmd, arguments, options)
    _, out, _ = cmdx(cmd, *arguments, **options)
    dd("cmdout: out:", out)
    return out


def cmdx(cmd, *arguments, **options):
    """
    Alias to k3proc.command() with ``check=True``.

    Returns:
        (int, list, list): exit code, stdout and stderr in lines of str.
    """
    dd("cmdx:", cmd, arguments, options)
    ddstack()

    options["check"] = True
    code, out, err = command(cmd, *arguments, **options)
    out = out.splitlines()
    err = err.splitlines()
    return code, out, err


def cmdtty(cmd, *arguments, **options):
    """
    Alias to k3proc.command() with ``check=True`` ``tty=True``.
    As if the command is run in a tty.

    Returns:
        (int, list, list): exit code, stdout and stderr in lines of str.
    """

    dd("cmdtty:", cmd, arguments, options)
    options["tty"] = True
    return cmdx(cmd, *arguments, **options)


def cmdpass(cmd, *arguments, **options):
    """
    Alias to k3proc.command() with ``check=True`` ``capture=False``.
    It just passes stdout and stderr to calling process.

    Returns:
        (int, list, list): exit code and empty stdout and stderr.
    """
    # interactive mode, delegate stdin to sub proc
    dd("cmdpass:", cmd, arguments, options)
    options["capture"] = False
    return cmdx(cmd, *arguments, **options)


def parse_flag(*flags):
    """
    Convert short form flag into tuple form, e.g.:
    parse_flag('x0') output: ('raise', 'oneline')

    '-x' will remove flag 'x'.
    parse_flag('x0-x') output ('online', )

    parse_flag(['raise', 'oneline', '-raise']) outputs ('oneline', )

    parse_flag(['raise', 'oneline', '-raise'], 't') outputs ('oneline', 'tty', )

    """

    expanded = []
    for flag in flags:
        f = expand_flag(flag)
        expanded.extend(f)

    #  reduce

    res = {}
    for key in expanded:
        if key.startswith("-"):
            key = key[1:]
            if key in res:
                del res[key]
        else:
            res[key] = True

    flag = tuple(res.keys())

    return flag


def expand_flag(flag):
    # expand abbreviations:
    # x  ->  raise
    # -x -> -raise

    mp = {
        "x": "raise",
        "t": "tty",
        "n": "none",
        "p": "pass",
        "o": "stdout",
        "0": "oneline",
    }

    if isinstance(flag, str):
        res = []
        buf = ""

        for c in flag:
            if c == "-":
                buf += c
                continue
            else:
                key = buf + mp[c]
                buf = ""

                res.append(key)

        flag = tuple(res)
    return flag



"""
k3handy is collection of mostly used  utilities.
"""

from importlib.metadata import version

__version__ = version("k3handy")

import os
import sys
import logging

from . import path

from k3fs import fread
from k3fs import fwrite
from k3fs import ls_dirs
from k3fs import ls_files
from k3fs import makedirs
from k3fs import remove
from k3proc import command
from k3proc import CalledProcessError
from k3proc import TimeoutExpired
from k3str import to_bytes

from .path import pabs
from .path import pjoin
from .path import prebase

from .cmd import cmd0
from .cmd import cmdf
from .cmd import cmdout
from .cmd import cmdpass
from .cmd import cmdtty
from .cmd import cmdx
from .cmd import parse_flag

from .cmd import dd
from .cmd import ddstack


__all__ = [
    # from k3fs
    "fread",
    "fwrite",
    "ls_dirs",
    "ls_files",
    "makedirs",
    "remove",
    # from k3proc
    "command",
    "CalledProcessError",
    "TimeoutExpired",
    # from k3str
    "to_bytes",
    # from .path
    "path",
    "pabs",
    "pjoin",
    "prebase",
    # from .cmd
    "cmd0",
    "cmdf",
    "cmdout",
    "cmdpass",
    "cmdtty",
    "cmdx",
    "parse_flag",
    "dd",
    "ddstack",
    # local
    "display",
]

logger = logging.getLogger(__name__)

#  Since 3.8 there is a stacklevel argument
ddstack_kwarg = {}
if sys.version_info.major == 3 and sys.version_info.minor >= 8:
    ddstack_kwarg = {"stacklevel": 2}


def display(stdout, stderr=None):
    """
    Output to stdout and stderr.
    - ``display(1, "foo")`` write to stdout.
    - ``display(1, ["foo", "bar"])`` write multilines to stdout.
    - ``display(1, ("foo", "bar"))`` write multilines to stdout.
    - ``display(("foo", "bar"), ["woo"])`` write multilines to stdout and stderr.
    - ``display(None, ["woo"])`` write multilines to stderr.

    """

    if isinstance(stdout, int):
        fd = stdout
        line = stderr

        if isinstance(line, (list, tuple)):
            lines = line
            for ln in lines:
                display(fd, ln)
            return

        os.write(fd, to_bytes(line))
        os.write(fd, b"\n")
        return

    if stdout is not None:
        display(1, stdout)

    if stderr is not None:
        display(2, stderr)



import unittest

import k3ut
import k3handy

dd = k3ut.dd


class TestHandyFs(unittest.TestCase):
    def test_found(self):
        _ = [
            k3handy.fread,
            k3handy.fwrite,
            k3handy.ls_dirs,
            k3handy.ls_files,
            k3handy.makedirs,
            k3handy.remove,
        ]



import unittest

import k3ut
import k3handy

dd = k3ut.dd


class TestHandyCmd(unittest.TestCase):
    def test_parse_flag(self):
        cases = [
            ([""], ()),
            (["x"], ("raise",)),
            (["t"], ("tty",)),
            (["n"], ("none",)),
            (["p"], ("pass",)),
            (["o"], ("stdout",)),
            (["0"], ("oneline",)),
            (
                ["x0"],
                (
                    "raise",
                    "oneline",
                ),
            ),
            (["x0-x"], ("oneline",)),
            (
                ["x0-xx"],
                (
                    "oneline",
                    "raise",
                ),
            ),
            (
                ["x0", "-xx"],
                (
                    "oneline",
                    "raise",
                ),
            ),
            (
                ["x0", ""],
                (
                    "raise",
                    "oneline",
                ),
            ),
            (["x0", ["-oneline"]], ("raise",)),
            ([("raise", "oneline", "-raise")], ("oneline",)),
        ]

        for c in cases:
            flags, want = c
            got = k3handy.parse_flag(*flags)
            self.assertEqual(want, got)

        with self.assertRaises(KeyError):
            k3handy.parse_flag("q")

    def test_cmdf(self):
        for f in ("0", ("oneline",)):
            got = k3handy.cmdf("python", "-c", 'print("a"); print("b")', flag=f)
            self.assertEqual("a", got)

            #  no output

            got = k3handy.cmdf("python", "-c", "", flag=f)
            self.assertEqual("", got)

            # not raise without 'x'
            k3handy.cmdf(
                "python",
                "-c",
                "import sys; sys.exit(5)",
                flag=f,
            )

        # return None if error

        for f in ("n0", ["none", "oneline"]):
            got = k3handy.cmdf(
                "python",
                "-c",
                "import sys; sys.exit(5)",
                flag=f,
            )
            self.assertEqual(None, got)

        #  raise with 'x'

        for f in ("x0", ["raise", "oneline"]):
            self.assertRaises(
                k3handy.CalledProcessError,
                k3handy.cmdf,
                "python",
                "-c",
                "import sys; sys.exit(5)",
                flag=f,
            )

        self.assertRaises(
            k3handy.CalledProcessError,
            k3handy.cmdx,
            "python",
            "-c",
            "import sys; sys.exit(5)",
        )

        # out
        for f in ("o", ["stdout"]):
            got = k3handy.cmdf(
                "python",
                "-c",
                'print("a"); print("b")',
                flag=f,
            )
            self.assertEqual(["a", "b"], got)

        for f in ("xo", ["raise", "stdout"]):
            self.assertRaises(
                k3handy.CalledProcessError,
                k3handy.cmdf,
                "python",
                "-c",
                "import sys; sys.exit(5)",
                flag=f,
            )

        # tty

        for f in ("t", ["tty"]):
            returncode, out, err = k3handy.cmdf(
                "python",
                "-c",
                "import sys; print(sys.stdout.isatty())",
                flag=f,
            )

            dd("out:", out)
            self.assertEqual(["True"], out)

        returncode, out, err = k3handy.cmdtty(
            "python",
            "-c",
            "import sys; print(sys.stdout.isatty())",
        )
        dd("out:", out)
        self.assertEqual(["True"], out)

        # input

        read_stdin_in_subproc = """
import k3handy;
k3handy.cmdf(
'python', '-c', 'import sys; print(sys.stdin.read())',
flag='p'
)
        """

        returncode, out, err = k3handy.cmdx(
            "python",
            "-c",
            read_stdin_in_subproc,
            input="123",
        )

        dd("out:", out)
        self.assertEqual(["123"], out)

    def test_cmd0(self):
        got = k3handy.cmd0(
            "python",
            "-c",
            'print("a"); print("b")',
        )
        self.assertEqual("a", got)

        #  no output

        got = k3handy.cmd0(
            "python",
            "-c",
            "",
        )
        self.assertEqual("", got)

        #  failure to exception

        self.assertRaises(
            k3handy.CalledProcessError,
            k3handy.cmd0,
            "python",
            "-c",
            "import sys; sys.exit(5)",
        )

    def test_cmdout(self):
        got = k3handy.cmdout(
            "python",
            "-c",
            'print("a"); print("b")',
        )
        self.assertEqual(["a", "b"], got)

        self.assertRaises(
            k3handy.CalledProcessError,
            k3handy.cmdout,
            "python",
            "-c",
            "import sys; sys.exit(5)",
        )

    def test_cmdx(self):
        got = k3handy.cmdx(
            "python",
            "-c",
            'print("a"); print("b")',
        )
        self.assertEqual((0, ["a", "b"], []), got)

        self.assertRaises(
            k3handy.CalledProcessError,
            k3handy.cmdx,
            "python",
            "-c",
            "import sys; sys.exit(5)",
        )

    def test_cmdtty(self):
        returncode, out, err = k3handy.cmdtty(
            "python",
            "-c",
            "import sys; print(sys.stdout.isatty())",
        )

        dd("returncode:", returncode)
        dd("out:", out)
        dd("err:", err)

        self.assertEqual(0, returncode)
        self.assertEqual(["True"], out)
        self.assertEqual([], err)

    def test_cmdpass(self):
        read_stdin_in_subproc = """
import k3handy;
k3handy.cmdpass(
'python', '-c', 'import sys; print(sys.stdin.read())',
)
        """

        returncode, out, err = k3handy.cmdx(
            "python",
            "-c",
            read_stdin_in_subproc,
            input="123",
        )

        dd("returncode:", returncode)
        dd("out:", out)
        dd("err:", err)

        self.assertEqual(0, returncode)
        self.assertEqual(["123"], out)



import unittest

import k3ut
import k3handy

dd = k3ut.dd


class TestHandyLogging(unittest.TestCase):
    def test_dd(self):
        script = """
import sys;
import logging;
import k3handy;
logging.basicConfig(stream=sys.stdout, level=logging.{level});
k3handy.dd("123");
"""
        got = k3handy.cmd0("python", "-c", script.format(level="DEBUG"))
        # 3.8.6 does not output ".cmd"
        # 3.8.10 does    output ".cmd"
        self.assertRegex(got, r"DEBUG:k3handy(\.cmd)?:123")

        got = k3handy.cmd0("python", "-c", script.format(level="INFO"))
        self.assertEqual("", got)

    def test_ddstack(self):
        script = """
import sys;
import logging;
import k3handy;
logging.basicConfig(stream=sys.stdout, level=logging.{level});
def foo(): k3handy.ddstack("123");
foo()
"""
        got = k3handy.cmdout("python", "-c", script.format(level="DEBUG"))

        # 3.8.6 does not output ".cmd"
        # 3.8.10 does    output ".cmd"
        self.assertRegex(got[0], r"DEBUG:k3handy(\.cmd)?:stack: 6 foo ")
        self.assertRegex(got[1], r"DEBUG:k3handy(\.cmd)?:stack: 7 <module> ")

        got = k3handy.cmdout("python", "-c", script.format(level="INFO"))
        self.assertEqual([], got)


class TestHandyDisplay(unittest.TestCase):
    def test_display(self):
        for cmd, want in (
            ('display(1, "foo")', (["foo"], [])),
            ('display(2, "foo")', ([], ["foo"])),
            ('display(2, ["foo", "bar"])', ([], ["foo", "bar"])),
            ('display(None, ["foo", "bar"])', ([], ["foo", "bar"])),
            ('display(["foo", "bar"], None)', (["foo", "bar"], [])),
            ('display(["foo", "bar"])', (["foo", "bar"], [])),
            ('display(["foo", "bar"], ["woo"])', (["foo", "bar"], ["woo"])),
            ('display("foo", "bar")', (["foo"], ["bar"])),
        ):
            _, out, err = k3handy.cmdx(
                "python",
                "-c",
                "import k3handy; k3handy.{}".format(cmd),
            )

            self.assertEqual(want, (out, err))



import os


def pjoin(*args):
    """
    Alias to os.path.join()

    >>> pjoin("a", "b")
    'a/b'

    >>> pjoin("a", "b", "c")
    'a/b/c'

    """
    return os.path.join(*args)


def pabs(*args):
    """
    Alias to os.path.abspath() and os.path.normpath()

    >>> pabs('a').startswith(os.path.abspath("."))
    True

    >>> pabs('a')[-2:]
    '/a'


    """
    p = os.path.abspath(pjoin(*args))
    return os.path.normpath(p)


def prebase(base, *pseg):
    """
    Rebaase path pseg on to base and returns the absolute path.

    >>> prebase("/a", "b")
    '/a/b'

    >>> prebase("/a", "b", "c")
    '/a/b/c'

    >>> prebase("/a", "/b")
    '/b'

    >>> prebase("/a", "b/../c")
    '/a/c'

    >>> prebase("/a")
    '/a'

    >>> repr(prebase("a", None))
    'None'

    >>> prebase(None, "/b")
    '/b'

    Args:

        base: base path, aka the left part.

        *pseg: segments to append to ``base``.

    Returns:
        str: the path joined
    """

    for p in pseg:
        if p is None:
            return None

        if os.path.isabs(p):
            base = p

        if base is None:
            base = pabs(p)

        base = pabs(base, p)

    return base
